import logging
import locale
import platform
import traceback
from typing import Optional, List
from datetime import datetime
from enum import Enum

#
# NB: for schemas refer to location below
# https://github.com/Microsoft/ApplicationInsights-Home/blob/master/EndpointSpecs/Schemas/Bond/
#
# NB: for tags keys refer to documentation in location below
# https://github.com/Microsoft/ApplicationInsights-Home/blob/master/EndpointSpecs/Schemas/Docs/ContextTagKeys.md


class Operation:
    __slots__ = ('id', 'name')

    def __init__(self, _id: str, name: str):
        self.id = _id
        self.name = name

    def to_dict(self):
        return {
            'ai.operation.id': self.id,
            'ai.operation.name': self.name
        }


class Session:
    __slots__ = ('id',)

    def __init__(self, _id: str):
        """Session ID - the instance of the user's interaction with the app."""
        self.id = _id

    def to_dict(self):
        return {
            'ai.session.id': self.id
        }


class User:
    __slots__ = ('account_id', 'id', 'session_id')

    def __init__(self, account_id: str, _id: str, session_id: str):
        self.account_id = account_id
        self.id = _id  # Anonymous user id. ???!
        self.session_id = session_id

    def to_dict(self):
        data = {}

        if self.account_id:
            data['ai.user.accountId'] = self.account_id

        if self.session_id:
            data['ai.session.id'] = self.session_id

        if self.id:
            data['ai.user.id'] = self.id
        return data


class LoggingDevice:
    """Information about the device that is logging information to Application Insights"""

    __slots__ = ('id', 'type_name', 'os_version', 'locale')

    def __init__(self, type_name=None):
        self.id = platform.node()
        self.type_name = type_name or 'PC'
        self.os_version = platform.version()
        self.locale = locale.getdefaultlocale()[0]

    def to_dict(self):
        return {
            'ai.device.id': self.id,
            'ai.device.locale': self.locale,
            'ai.device.osVersion': self.os_version,
            'ai.device.type': self.type_name
        }


class Application:
    __slots__ = ('version',)

    def __init__(self, version: str):
        self.version = version

    def to_dict(self):
        return {
            'ai.application.ver': self.version
        }


class Context:
    """Telemetry client context, common to all collected telemetries"""
    __slots__ = ('application', 'device')

    def __init__(self,
                 application: Optional[Application]=None,
                 device: Optional[LoggingDevice]=None):
        self.application = application
        self.device = device or LoggingDevice()


class TraceSeverity:
    verbose = 0

    information = 1

    warning = 2

    error = 3

    critical = 4


class TraceData:
    """
    NB: this class is called 'trace' in Azure Portal and anywhere else;
    somebody in Microsoft must have disliked the choice of using such generic names like 'Message'
    """
    envelope_type_name = 'Microsoft.ApplicationInsights.Message'

    data_type_name = 'MessageData'

    logging_levels = {
        'DEBUG': 0,
        'INFO': 1,
        'WARNING': 2,
        'ERROR': 3,
        'CRITICAL': 4,
        logging.DEBUG: 0,
        logging.INFO: 1,
        logging.WARNING: 2,
        logging.ERROR: 3,
        logging.CRITICAL: 4
    }

    __slots__ = ('message', 'properties', 'severity')

    def __init__(self, message: str, properties: Optional[dict]=None, severity: int=1):
        self.message = message
        self.properties = properties
        self.severity = severity

    def to_dict(self):
        return {
            'message': self.message,
            'properties': self.properties,
            'severityLevel': self.severity
        }


class EventData:
    envelope_type_name = 'Microsoft.ApplicationInsights.Event'

    data_type_name = 'EventData'

    __slots__ = ('name', 'properties', 'measurements')

    def __init__(self,
                 name: str,
                 properties: dict,
                 measurements: dict):
        self.name = name
        self.properties = properties
        self.measurements = measurements

    def to_dict(self):
        return {
            'name': self.name,
            'properties': self.properties,
            'measurements': self.measurements
        }


class StackFrame:

    __slots__ = ('level',
                 'method',
                 'module',
                 'file_name',
                 'line')

    def __init__(self,
                 level: int,
                 method: str,
                 module: str,
                 file_name: str,
                 line: int):
        self.level = level
        self.method = method
        # in .NET there are assemblies, in Python not such a thing, but we can log modules
        self.module = module
        self.file_name = file_name
        self.line = line

    def to_dict(self):
        return {
            'level': self.level,
            'method': self.method,
            'assembly': self.module,
            'fileName': self.file_name,
            'line': self.line
        }


class ExceptionDetails:

    __slots__ = ('id',
                 'outer_id',
                 'message',
                 'type_name',
                 'has_full_stack',
                 'stack',
                 'texts')

    def __init__(self,
                 _id: int,
                 type_name: str,
                 message: str,
                 stack: List[StackFrame],
                 texts: List[str]=None,
                 outer_id: int = 0,
                 has_full_stack: bool = True):
        self.id = _id
        self.outer_id = outer_id
        self.type_name = type_name
        self.message = message
        self.has_full_stack = has_full_stack
        self.stack = stack
        self.texts = texts

    @classmethod
    def from_exception(cls, type, value, tb, skip_frames=0):
        _id = 1
        outer_id = 0
        type_name = type.__name__
        message = str(value)

        stack = []
        texts = []
        index = 0
        for tb_frame_file, tb_frame_line, tb_frame_function, tb_frame_text in traceback.extract_tb(tb):
            if skip_frames:
                skip_frames -= 1
                continue

            texts.append(tb_frame_text)

            stack.append(StackFrame(index,
                                    tb_frame_function,
                                    ' ',
                                    tb_frame_file,
                                    tb_frame_line))
            index += 1

        stack.reverse()
        texts.reverse()
        return cls(_id,
                   type_name,
                   message,
                   stack,
                   texts,
                   outer_id,
                   True)

    def to_dict(self):
        return {
            'id': self.id,
            'outerId': self.outer_id,
            'typeName': self.type_name,
            'message': self.message,
            'hasFullStack': self.has_full_stack,
            'parsedStack': self.stack
        }


class ExceptionData:
    envelope_type_name = 'Microsoft.ApplicationInsights.Exception'

    data_type_name = 'ExceptionData'

    __slots__ = ('exceptions', 'properties', 'measurements')

    def __init__(self,
                 exceptions: List[ExceptionDetails],
                 properties: Optional[dict] = None,
                 measurements: Optional[dict] = None):
        self.exceptions = exceptions
        self.properties = properties
        self.measurements = measurements

    def to_dict(self):
        data = {
            'handledAt': 'UserCode',
            'exceptions': self.exceptions
        }

        for name in ('properties', 'measurements'):
            v = getattr(self, name)
            if v:
                data[name] = v

        return data


class DataPointKind(Enum):
    Measurement = 1,
    Aggregation = 2

    def __int__(self):
        return self.value[0]

class DataPointTypes:
    Measurement = 1
    Aggregation = 2


class DataPoint:
    __slots__ = ('name',
                 'kind',
                 'value',
                 'count',
                 'min',
                 'max',
                 'std_dev')

    def __init__(self,
                 name: str,
                 value: float,
                 kind: DataPointKind=None,
                 count: Optional[int]=None,
                 min: Optional[float]=None,
                 max: Optional[float]=None,
                 std_dev: Optional[float]=None):
        # NB
        # https://github.com/Microsoft/ApplicationInsights-Home/blob/master/EndpointSpecs/Schemas/Bond/DataPoint.bond
        if kind is None:
            kind = DataPointKind.Measurement

        self.name = name
        self.kind = kind
        self.value = value
        self.count = count
        self.min = min
        self.max = max
        self.std_dev = std_dev

    def to_dict(self):
        return {x: v for x, v in {
            'name': self.name,
            'value': self.value,
            'kind': int(self.kind),
            'count': self.count,
            'min': self.min,
            'max': self.max,
            'stdDev': self.std_dev
        }.items() if v is not None}


class MetricData:
    envelope_type_name = 'Microsoft.ApplicationInsights.Metric'

    data_type_name = 'MetricData'

    __slots__ = ('item', 'properties')

    def __init__(self, item: DataPoint, properties):
        self.item = item
        self.properties = properties

    def to_dict(self):
        # Metrics document schema defines an array of metrics.
        # However Application Insights only supports one element in this array. (D'oh!)
        return {
            'metrics': [self.item],
            'properties': self.properties
        }


class RequestData:
    envelope_type_name = 'Microsoft.ApplicationInsights.Request'

    data_type_name = 'RequestData'

    __slots__ = ('id',
                 'url',
                 'name',
                 'start_time',
                 'duration',
                 'response_code',
                 'success',
                 'http_method',
                 'properties',
                 'measurements')

    def __init__(self,
                 _id: str,
                 name: str,
                 http_method: str,
                 url: str,
                 response_code: str,
                 success: bool,
                 start_time: datetime,
                 duration: int,
                 properties: Optional[dict],
                 measurements: Optional[dict]):
        self.id = _id
        self.url = url
        self.name = name
        self.start_time = start_time
        self.duration = duration
        self.response_code = int(response_code)
        self.success = success
        self.http_method = http_method
        self.properties = properties
        self.measurements = measurements

    @staticmethod
    def format_duration(duration: int):
        #
        # Application Insights expect this format: hh:mm:ss.fff
        #
        if not duration:
            duration = 0

        if duration < 0:
            raise ValueError('request duration cannot be negative')

        parts = []
        for multiplier in [1000, 60, 60, 24]:
            parts.append(duration % multiplier)
            duration //= multiplier
        parts.reverse()
        formatted_duration = '%02d:%02d:%02d.%03d' % tuple(parts)

        if duration:
            formatted_duration = '%d.%s' % (duration, formatted_duration)

        return formatted_duration

    def to_dict(self):
        data = {
            'id': self.id,
            'name': self.name,
            'startTime': self.start_time,
            'duration': self.format_duration(self.duration),
            'responseCode': str(self.response_code),
            'success': self.success,
            'httpMethod': self.http_method,
            'url': self.url
        }

        if self.properties:
            data['properties'] = self.properties

        if self.measurements:
            data['measurements'] = self.measurements

        return data


class Envelope:

    __slots__ = ('data',
                 'name',
                 'data_type_name',
                 'time',
                 'instrumentation_key',
                 'tags')

    def __init__(self,
                 instrumentation_key: str,
                 data,
                 tags: dict):
        self.data = data
        self.name = data.envelope_type_name
        self.data_type_name = data.data_type_name
        self.time = datetime.utcnow()
        self.instrumentation_key = instrumentation_key
        self.tags = tags

    def __repr__(self):
        return f'<Envelope {self.data_type_name}>'

    def to_dict(self):
        data = self.data.to_dict()
        data['ver'] = 2
        return {
            'ver': 1,
            'name': self.name,
            'time': self.time,
            'sampleRate': 100.00,
            'iKey': self.instrumentation_key,
            'tags': self.tags,
            'data': {
                'baseType': self.data_type_name,
                'baseData': data
            }
        }


import unittest


def cases(*cases):
    def decorator(f):
        f.cases = cases
        return f
    return decorator


def _apply(func, args):
    """Return a function with args applied after first argument"""
    def wrapped(self):
        return func(self, *args)
    return wrapped


class TheoryMeta(type):
    """Metaclass that replaces test methods with multiple methods for each test case"""
    def __new__(meta, name, bases, attrs):
        newattrs = {}

        for name, value in attrs.items():
            if not name.startswith('test') or not callable(value):
                newattrs[name] = value
                continue

            if not hasattr(value, 'cases'):
                newattrs[name] = value
                continue

            cases = value.cases

            for n, args in enumerate(cases):
                test_name = '%s_%d' % (name, n)
                newattrs[test_name] = _apply(value, args)

        return super().__new__(meta, name, bases, newattrs)


class Theory(unittest.TestCase, metaclass=TheoryMeta):
    pass



import sys
import uuid
import asyncio
from datetime import datetime
from typing import Optional, Union
from .channel.abstractions import TelemetryChannel
from .entities import (Application,
                       LoggingDevice,
                       Context,
                       Operation,
                       Session,
                       User,
                       Envelope,
                       EventData,
                       TraceData,
                       MetricData,
                       DataPoint,
                       DataPointKind,
                       RequestData,
                       ExceptionData,
                       ExceptionDetails)
from .utils import require_params


COMMON_TAGS = {
    'ai.internal.sdkVersion': 'asynpy3:0.0.1'
}


class AsyncTelemetryClient:
    """Azure Application Insights client using asyncio"""
    __slots__ = ('instrumentation_key',
                 '_context',
                 '_channel')

    def __init__(self,
                 instrumentation_key: str,
                 channel: TelemetryChannel,
                 application: Optional[Application]=None,
                 device: Optional[LoggingDevice]=None):
        require_params(instrumentation_key=instrumentation_key,
                       channel=channel)
        self.instrumentation_key = instrumentation_key
        self._context = Context(application, device)
        self._channel = channel

    def handle_unhandled_exceptions(self, loop=None):
        """
        Registers an handler to log all unhandled exceptions with this telemetry client.

        :param loop: loop to execute asynchronous exception logging, if not given asyncio.get_event_loop() is used
        """
        current_excepthook = sys.excepthook

        if loop is None:
            # use implicit loop
            loop = asyncio.get_event_loop()

        async def log_exception_async(type, value, traceback):
            # NB: flush is needed because in case of unhandled exception the whole process might fall
            await self.track_exception(type, value, traceback)
            await self.flush()

        def local_excepthook(type, value, traceback):
            #
            loop.run_until_complete(log_exception_async(type, value, traceback))

            # call the original method
            current_excepthook(type, value, traceback)

        sys.excepthook = local_excepthook
        pass

    async def push(self, data):
        await self._channel.put(data)

    async def flush(self):
        await self._channel.flush()

    async def __aenter__(self):
        return self

    async def __aexit__(self,
                        exc_type,
                        exc_val,
                        exc_tb):
        await self.dispose()

    def get_tags(self,
                 operation: Optional[Operation]=None,
                 session: Optional[Session]=None,
                 user: Optional[User]=None
                 ) -> dict:
        """
        Returns meta tags to be included in any log.
        ref.
        https://github.com/Microsoft/ApplicationInsights-Home/blob/master/EndpointSpecs/Schemas/Docs/ContextTagKeys.md

        :param operation: optional operation data to log.
        :param session: optional session data to log.
        :param user: optional user data to log.
        :return:
        """
        tags = self._context.device.to_dict()

        if self._context.application:
            tags.update(self._context.application.to_dict())

        if operation:
            tags.update(operation.to_dict())

        if user:
            tags.update(user.to_dict())

        if session:
            tags.update(session.to_dict())

        tags.update(COMMON_TAGS)
        return tags

    async def track_event(self,
                          name,
                          properties=None,
                          measurements=None,
                          *,
                          operation: Optional[Operation] = None,
                          session: Optional[Session] = None,
                          user: Optional[User] = None):
        """
        Logs a single event.

        :param name: event name
        :param properties: event properties
        :param measurements: event measurements
        :param operation: optional operation tags to log
        :param session: optional session tags to log
        :param user: optional user tags to log
        :return:
        """
        data = Envelope(self.instrumentation_key,
                        EventData(name, properties, measurements),
                        self.get_tags(operation, session, user))

        await self.push(data)

    async def track_exception(self,
                              type=None,
                              value=None,
                              tb=None,
                              properties=None,
                              measurements=None,
                              *,
                              operation: Optional[Operation] = None,
                              session: Optional[Session] = None,
                              user: Optional[User] = None,
                              skip_frames: Optional[int] = None):
        """
        Tracks a single exception, trying to obtain information from sys.exc_info if type, value and traceback
        are not given.

        :param type: exception type
        :param value: exception value
        :param tb: traceback object
        :param properties: optional properties
        :param measurements: optional measurements
        :param operation: optional operation tags to log
        :param session: optional session tags to log
        :param user: optional user tags to log
        :param skip_frames: optional number of stack frames to be skipped from log
        """
        if not type or not value or not tb:
            type, value, tb = sys.exc_info()

        if not type or not value or not tb:
            try:
                raise Exception('Technical exception')
            except:
                type, value, tb = sys.exc_info()

        details = ExceptionDetails.from_exception(type, value, tb, skip_frames)

        # Python trace back gives us also portions of source code; useful information
        # the official application insights sdk for Python discards it, but it can be stored
        # in custom data like done here:
        index = 0
        text_portions = {}
        for text in details.texts:
            text_portions[f'code_{index}'] = text
            index += 1

        if properties:
            properties.update(text_portions)
        else:
            properties = text_portions

        data = Envelope(self.instrumentation_key,
                        ExceptionData([details],
                                      properties,
                                      measurements),
                        self.get_tags(operation, session, user))

        await self.push(data)

    async def track_trace(self,
                          name,
                          properties=None,
                          severity: int=1,
                          *,
                          operation: Optional[Operation] = None,
                          session: Optional[Session] = None,
                          user: Optional[User] = None):
        """
        Logs a single trace.

        :param name: trace name
        :param properties: trace properties
        :param severity: severity level, refer to entities.TraceSeverity class
        :param operation: optional operation tags to log
        :param session: optional session tags to log
        :param user: optional user tags to log
        """
        data = Envelope(self.instrumentation_key,
                        TraceData(name, properties, severity),
                        self.get_tags(operation, session, user))

        await self.push(data)

    async def track_metric(self,
                           name: str,
                           value: float,
                           kind: DataPointKind = None,
                           count: Optional[int] = None,
                           min: Optional[float] = None,
                           max: Optional[float] = None,
                           std_dev: Optional[float] = None,
                           properties: Optional[dict] = None,
                           *,
                           operation: Optional[Operation] = None,
                           session: Optional[Session] = None,
                           user: Optional[User] = None
                           ):
        """
        Logs a single metric; measurement or aggregation data.

        :param name: metric name
        :param value: metric value
        :param kind: metric kind (Measurement | Aggregation)
        :param count: optional count
        :param min: optional minimum value
        :param max: optional maximum value
        :param std_dev: optional standard deviation
        :param properties: optional extra properties to log
        :param operation: optional operation tags to log
        :param session: optional session tags to log
        :param user: optional user tags to log
        :return:
        """
        item = DataPoint(name, value, kind, count, min, max, std_dev)

        data = Envelope(self.instrumentation_key,
                        MetricData(item, properties),
                        self.get_tags(operation, session, user))

        await self.push(data)

    async def track_request(self,
                            _id: str,
                            name: str,
                            url: str,
                            success: bool,
                            start_time: datetime=None,
                            duration:Optional[int]=None,
                            response_code:Union[str, int]=None,
                            http_method: str=None,
                            properties: Optional[dict]=None,
                            measurements: Optional[dict]=None,
                            *,
                            operation: Optional[Operation] = None,
                            session: Optional[Session] = None,
                            user: Optional[User] = None
                            ):
        """
        Logs a single HTTP request captured by a web application.

        :param _id: telemetry id assigned to the request
        :param name: the name of request to log
        :param url: actual URL
        :param success: whether the request was processed with success
        :param start_time: request start time when received by the app server
        :param duration: the number of milliseconds it took to process the request
        :param response_code: response code
        :param http_method: request HTTP method
        :param properties: set of custom properties to store
        :param measurements: set of custom measurements to store
        :return:
        """
        if not start_time:
            start_time = datetime.utcnow()

        request_id = _id or str(uuid.uuid4())

        if not operation:
            # in this case, operation tags can be configured automatically if not specified
            # in caller method
            operation = Operation(request_id, f'{http_method} {name}')

        data = Envelope(self.instrumentation_key,
                        RequestData(request_id,
                                    name,
                                    http_method,
                                    url,
                                    response_code,
                                    success,
                                    start_time,
                                    duration or 0,
                                    properties,
                                    measurements),
                        self.get_tags(operation, session, user))

        await self.push(data)

    async def dispose(self):
        try:
            await self._channel.flush()
        finally:
            await self._channel.dispose()
        return self



"""
* aiohttp integration with Azure Application Insights
*
* https://github.com/RobertoPrevato/asynapplicationinsights
*
* Copyright 2018, Roberto Prevato
* https://robertoprevato.github.io
*
* Licensed under the MIT license:
* http://www.opensource.org/licenses/MIT
"""
import uuid
import time
from typing import Optional, Callable
from aiohttp import web, ClientSession
from aiohttp.web_exceptions import HTTPException
from datetime import datetime
from .telemetry import AsyncTelemetryClient
from .entities import Application, LoggingDevice
from .channel.aiohttpchannel import AiohttpTelemetryChannel


def default_time_getter():
    return datetime.utcnow()


# success codes from server perspective!
client_errors_request_handling_success = {401, 403, 404, 405}


def default_is_success_request(status):
    if status < 400:
        return True

    if status in client_errors_request_handling_success:
        # failed from client perspective, but successfully handled by server perspective!
        # such as unauthorized, forbidden, not found
        return True

    return False


def get_request_name(request):
    ascii_encodable_path = request.path.encode('ascii', 'backslashreplace') \
        .decode('ascii')
    return ascii_encodable_path



def use_application_insights(app: web.Application,
                             instrumentation_key: str,
                             app_metadata: Optional[Application] = None,
                             logging_device: Optional[LoggingDevice] = None,
                             time_getter: Optional[Callable] = None,
                             user_getter: Optional[Callable] = None,
                             is_success_request: Optional[Callable] = None,
                             is_handled_exception: Optional[Callable] = None,
                             requests_filter: Optional[Callable] = None,
                             client_session: ClientSession=None,
                             loop=None):
    """
    Integrates asynchronous client for Azure Application Insights into an aiohttp application.

    :param app: aiohttp application
    :param instrumentation_key: application insights instrumentation key
    :param app_metadata: optional metadata about the application
    :param logging_device: optional metadata about the logging device; if not specified one is created with platform information
    :param time_getter: optional method to return current time for request, if not specified datetime.utcnow is used
    :param user_getter: optional method to obtain user metadata information from request
    :param is_success_request: optional method to determine whether a request was successful from server perspective
    :param is_handled_exception: optional method to determine whether an exception is handled and thrown intentionally
    :param requests_filter: optional method to filter requests from ai logging
    :param loop: optional asyncio loop, if not specified asyncio.get_event_loop is used
    :param client_session: optionally, an http client session for web requests
    :return:
    """
    if loop is None:
        loop = app.loop

    if not time_getter:
        time_getter = default_time_getter

    if not is_success_request:
        is_success_request = default_is_success_request

    client = AsyncTelemetryClient(instrumentation_key,
                                  AiohttpTelemetryChannel(loop, client_session),
                                  app_metadata,
                                  logging_device)

    # on clean up, dispose the client
    async def on_clean_up_dispose_ai_client(_):
        await client.track_event('Application_Stop')
        await client.dispose()

    app.on_cleanup.append(on_clean_up_dispose_ai_client)

    setattr(app, 'ai_client', client)

    @web.middleware
    async def application_insights_middleware(request, handler):

        if requests_filter and requests_filter(request):
            return await handler(request)

        start_datetime = time_getter()
        start = time.time()
        req_name = get_request_name(request)
        req_url = str(request.url)
        telemetry_id = str(uuid.uuid4())
        user_data = None

        request.telemetry_id = telemetry_id

        try:
            response = await handler(request)

            # restore user context if possible, this must happen here
            if user_getter:
                user_data = user_getter(request)

            elapsed = time.time() - start
            elapsed_ms = int(elapsed * 1000)

            success = is_success_request(response.status)
            await client.track_request(telemetry_id,
                                       req_name,
                                       req_url,
                                       success,
                                       start_datetime,
                                       elapsed_ms,
                                       response.status,
                                       request.method,
                                       user=user_data)
            return response
        except HTTPException as http_exception:
            elapsed = time.time() - start
            elapsed_ms = int(elapsed * 1000)

            # restore user context if possible, this must happen here
            if user_getter:
                user_data = user_getter(request)

            status = http_exception.status
            success = is_success_request(status)
            await client.track_request(telemetry_id,
                                       req_name,
                                       req_url,
                                       success,
                                       start_datetime,
                                       elapsed_ms,
                                       status,
                                       request.method,
                                       user=user_data)
            raise
        except Exception as exception:
            # log exception
            elapsed = time.time() - start
            elapsed_ms = int(elapsed * 1000)

            # restore user context if possible, this must happen here
            if user_getter:
                user_data = user_getter(request)

            if is_handled_exception:
                is_handled, status = is_handled_exception(exception)

                if is_handled:
                    success = is_success_request(status)
                    await client.track_request(telemetry_id,
                                               req_name,
                                               req_url,
                                               success,
                                               start_datetime,
                                               elapsed_ms,
                                               status,
                                               request.method,
                                               user=user_data)
                    raise

            status = 500
            success = False
            await client.track_request(telemetry_id,
                                       req_name,
                                       req_url,
                                       success,
                                       start_datetime,
                                       elapsed_ms,
                                       status,
                                       request.method,
                                       user=user_data)

            await client.track_exception(exception.__class__,
                                         exception,
                                         exception.__traceback__,
                                         user=user_data,
                                         skip_frames=1)
            raise

    app.middlewares.append(application_insights_middleware)

    return application_insights_middleware




from . import Theory, cases
from ..entities import RequestData


class TestEntities(Theory):

    @cases(
        (1000 * 16 + 150, '00:00:16.150'),
        (1000 * 16 + 15, '00:00:16.015'),
        (255, '00:00:00.255'),
        (1155, '00:00:01.155'),
        (50, '00:00:00.050'),
        (5, '00:00:00.005'),
    )
    def test_request_duration_formatting(self, value, expected_format):
        formatted = RequestData.format_duration(value)
        self.assertEqual(expected_format, formatted)




def exception_str(ex):
    return str(ex) or ex.__class__.__name__


class OperationFailed(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidOperation(Exception):
    """An exception risen in case of an operation that doesn't make sense in a certain context."""


class ArgumentNullException(ValueError):
    """An exception risen when a null or empty parameter is not acceptable."""
    def __init__(self, param_name):
        super().__init__(f'Parameter cannot be null or empty: `{param_name}`')


class InvalidArgument(Exception):
    pass



from asyncio import Queue, QueueEmpty
from abc import ABC, abstractmethod
from typing import List


class TelemetryChannel(ABC):

    def __init__(self):
        self._queue = Queue()
        self._max_length = 500

    def get(self):
        try:
            return self._queue.get_nowait()
        except QueueEmpty:
            return None

    async def put(self, item):
        if not item:
            return
        await self._queue.put(item)
        if self.should_flush():
            await self.flush()

    def should_flush(self) -> bool:
        return self._max_length <= self._queue.qsize()

    async def flush(self):
        data = []
        while True:
            item = self.get()
            if not item:
                break
            data.append(item)

        await self.send(data)

    @abstractmethod
    async def send(self, data: List):
        pass

    @abstractmethod
    async def dispose(self):
        pass



import uuid
import asyncio
import unittest
from datetime import datetime, timedelta
from .settings import instrumentation_key
from ..telemetry import AsyncTelemetryClient
from ..channel.aiohttpchannel import AiohttpTelemetryChannel
from random import randint


class TestAi(unittest.TestCase):

    def test_write_event(self):
        loop = asyncio.get_event_loop()

        async def go():
            client = AsyncTelemetryClient(instrumentation_key,
                                          AiohttpTelemetryChannel())

            await client.track_event('Example',
                                     measurements={
                                         'speed': 55.5
                                     })

            await client.track_trace('Foo', {'foo': 'power'}, 2)

            for _ in range(10):
                await asyncio.sleep(0.005)
                await client.track_metric('Something',
                                          randint(0, 200))

            await client.track_request(str(uuid.uuid4()),
                                       'Example',
                                       'http://localhost:44666/like/123456',
                                       True,
                                       datetime.utcnow() + timedelta(seconds=-10),
                                       duration=125,
                                       response_code=200,
                                       http_method='GET'
                                       )

            try:
                x = 100 / 0
            except:
                await client.track_exception()

            await client.flush()
            await client.dispose()

        loop.run_until_complete(go())



from ..exceptions import ArgumentNullException


def has_params(data, *args):
    """
    Validates required parameters against an object.

    :param data:
    :param args: required parameters
    :return:
    """
    if not data:
        return False
    for a in args:
        if not a in data:
            return False
        v = data[a]
        if not v or v.isspace():
            return False
    return True


def is_empty_string(value):
    return isinstance(value, str) and value.isspace()


def require_params(*args, **kwargs):
    for name, value in kwargs.items():
        if not value or is_empty_string(value):
            raise ArgumentNullException(name)

    for value in args:
        if not value or is_empty_string(value):
            raise ArgumentNullException(value)



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='asynapplicationinsights',
      version='0.0.3',
      description='Azure Application Insights client using asyncio',
      long_description=readme(),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
          'Operating System :: OS Independent',
          'Framework :: AsyncIO'
      ],
      url='https://github.com/RobertoPrevato/asynapplicationinsights',
      author='Roberto Prevato',
      author_email='roberto.prevato@gmail.com',
      keywords='asyncio aiohttp azure application insights telemetry',
      license='MIT',
      packages=['asynapplicationinsights',
                'asynapplicationinsights.channel',
                'asynapplicationinsights.tests',
                'asynapplicationinsights.utils'],
      install_requires=[
          'aiohttp',
      ],
      include_package_data=True,
      zip_safe=False)



import aiohttp
import asyncio
from typing import Optional, List
from .abstractions import TelemetryChannel
from ..exceptions import OperationFailed
from ..utils.json import friendly_dumps


class AiohttpTelemetryChannel(TelemetryChannel):

    def __init__(self,
                 loop:Optional[asyncio.AbstractEventLoop]=None,
                 client:Optional[aiohttp.ClientSession]=None,
                 endpoint:Optional[str]=None):
        super().__init__()

        dispose_client = True
        if client is None:
            if loop is None:
                loop = asyncio.get_event_loop()
            client = aiohttp.ClientSession(loop=loop)
        else:
            dispose_client = False

        if not endpoint:
            endpoint = 'https://dc.services.visualstudio.com/v2/track'

        self._dispose_client = dispose_client
        self._http_client = client
        self._endpoint = endpoint
        self._headers = {'Accept': 'application/json', 'Content-Type': 'application/json; charset=utf-8'}

    async def send(self, data: List):
        body = friendly_dumps(data)
        response = await self._http_client.post(self._endpoint,
                                                data=body.encode('utf8'),
                                                headers=self._headers)

        if response.status != 200:
            text = await response.text()
            raise OperationFailed(f'Response status does not indicate success: {response.status}; response body: {text}')

    async def dispose(self):
        # NB: the client is disposed only if it was instantiated
        if self._dispose_client:
            await self._http_client.close()



"""
This module defines a more user-friendly json encoder, supporting time objects and UUID
"""
import json
from datetime import time, date, datetime
from uuid import UUID


class FriendlyEncoder(json.JSONEncoder):

    def default(self, obj):
        if hasattr(obj, "to_dict"):
            return obj.to_dict()
        if isinstance(obj, time):
            return obj.strftime("%H:%M:%S")
        if isinstance(obj, datetime):
            return obj.isoformat() + 'Z'
        if isinstance(obj, date):
            return obj.strftime("%Y-%m-%d")
        if isinstance(obj, bytes):
            return obj.decode('utf8')
        if isinstance(obj, UUID):
            return str(obj)
        return json.JSONEncoder.default(self, obj)


def friendly_dumps(obj, skipkeys=False, ensure_ascii=False, check_circular=True,
        allow_nan=True, cls=None, indent=None, separators=None,
        default=None, sort_keys=False, **kw):
    if cls is None:
        cls = FriendlyEncoder
    return json.dumps(obj,
                      skipkeys=skipkeys,
                      ensure_ascii=ensure_ascii,
                      check_circular=check_circular,
                      allow_nan=allow_nan,
                      cls=cls,
                      indent=indent,
                      separators=separators,
                      default=default,
                      sort_keys=sort_keys,
                      **kw)
